### Fitting the linear models
### Evaluating the models
### Calculating variable importance


from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import train_test_split
from sklearn.model_selection import KFold, cross_val_score, cross_validate, GridSearchCV, PredefinedSplit, RandomizedSearchCV
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from pathlib import Path
from math import sqrt
from scipy import interpolate
from pprint import pprint
import matplotlib.pyplot as plt
import pandas as pd
import pickle
import os
import numpy as np
import sklearn

root = Path("my/path")

fold_files = sorted(list(root.glob("*")))
fold_files

folds = [
    pd.read_csv(root / f"fold{i+1}.csv", sep=";", decimal=",").drop(columns=['X', 'Y']) for i in range(5)
]

test = pd.read_csv(root / "test_set.csv", sep=";", decimal=",").drop(columns=['X', 'Y'])

my_variable = 'bio1' #The target variable



# ## Model fitting

# ### 1. RF - Random forest 


five_splits = [
     [fold.drop(columns=['bio1', 'bio2', 'bio3',
                               'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11',
                               'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18',
                               'bio19']),fold[my_variable]]
               for fold in folds
]

Xtest = test.drop(columns=['bio1', 'bio2', 'bio3','bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 
                        'bio11','bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18','bio19'])
ytest =  test[my_variable]  # Target variable for the test data



# Initialize the best_models and the search_models list
best_models = [] 
search_models = [] 

test_metrics = {
    "r2": [],
    "adjusted-r2": [],
    "rmse": [],
    "mae": [],
}


# Splits for training and validation data
train_val_splits = five_splits

# Concatenated features and target variable for training and validation
X_train_val = np.concatenate([x for x,_ in train_val_splits])
y_train_val = np.concatenate([y for _,y in train_val_splits]) 

# Start indices of each split
split_start_indices = np.cumsum([0] + [len(x) for x,_ in train_val_splits])
split_indices = np.concatenate([i*np.ones_like(y) for i, (_, y) in enumerate(train_val_splits)])  # split_indices[i] = 0 means sample i is part of fold0, and so on

# Predefined cross-validation split
cv = PredefinedSplit(test_fold=split_indices)

# Grid search with cross-validation
search_model = GridSearchCV(estimator=RandomForestRegressor(), param_grid={'n_estimators': [500, 600, 700]}, cv=cv, refit=True, verbose=4)  # Since refit=True, best model will be refit on the whole training+validation data

# Fit the model on training and validation data
search_model.fit(X_train_val, y_train_val)

# Make predictions on the test data
y_pred = search_model.predict(Xtest)

# Compute test metrics
test_metrics = {
    "r2": r2_score(ytest, y_pred),
    "adjusted-r2": 1 - ((1 - r2_score(ytest, y_pred)) * (len(ytest) - 1) / (len(ytest) - Xtest.shape[1] - 1)),
    "rmse": sqrt(mean_squared_error(ytest, y_pred)),
    "mae": mean_absolute_error(ytest, y_pred)
}

print(test_metrics)

# Then we repeat this procedure for all the response variables of interest


# ### 2. XGBOOST- Extreme Gradient BOOSTing

import xgboost as xgb

xbgr = xgb.XGBRegressor(verbosity=0, random_state=42)

test_metrics = {
    "r2": [],
    "adjusted-r2": [],
    "rmse": [],
    "mae": [],
}

best_models = [] # Initialize the best_models list
search_models = [] # Initialize the search_models list

Xtest = test.drop(columns=['bio1', 'bio2', 'bio3','bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 
                        'bio11','bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18','bio19'])
ytest =  test[my_variable]  


param_grid = {
        'silent': [False],
        'max_depth': [6, 10, 15, 20],
        'learning_rate': [0.01, 0.1, 1],
        'subsample': [0.6, 0.7],
        'colsample_bytree': [0.7, 0.8], 
        'colsample_bylevel': [0.5, 0.65, 0.8], 
        'min_child_weight': [0.5, 1.0],
        'gamma': [0, 0.5, 1.0],
        'reg_lambda': [0.1, 1.0, 50.0],
        'n_estimators': [200, 400, 600],
        'objective' : ['reg:squarederror']}


train_val_splits = five_splits

X_train_val = np.concatenate([x for x,_ in train_val_splits])
y_train_val = np.concatenate([y for _,y in train_val_splits])

split_start_indices = np.cumsum([0] + [len(x) for x,_ in train_val_splits])
split_indices = np.concatenate([ i*np.ones_like(y) for i, (_, y) in enumerate(train_val_splits) ]) # split_indices[i] = 0 means sample i is part of fold0, and so on

cv = PredefinedSplit(test_fold=split_indices)

search_model = RandomizedSearchCV(estimator=xbgr, 
                                  param_distributions=param_grid, 
                                  n_iter=10,
                                  verbose=4, 
                                  cv=cv,
                                  refit=True, # Since refit=True, best model will be refit on the whole training+validation data
                                  random_state=42)   

search_model.fit(X_train_val, y_train_val) 


# Make predictions on the test data
y_pred = search_model.predict(Xtest)

test_metrics["r2"].append(r2_score(ytest, y_pred))
test_metrics["rmse"].append(sqrt(mean_squared_error(ytest, y_pred)))
test_metrics["mae"].append(mean_absolute_error(ytest, y_pred))
test_metrics["adjusted-r2"].append(1 - ((1 - r2_score(ytest, y_pred)) * (len(ytest) - 1) / (len(ytest) - Xtest.shape[1] - 1)))
test_metrics

search_models.append(search_model.best_params_)


# Evaluate on test data and remember the best model

print(test_metrics)


# ### 3. EXTRA - EXTremely RAndomized trees

from sklearn.ensemble import ExtraTreesRegressor

etr = ExtraTreesRegressor()

test_metrics = {
    "r2": [],
    "adjusted-r2": [],
    "rmse": [],
    "mae": [],
}

best_models = [] # Initialize the best_models list
search_models = [] # Initialize the search_models list

Xtest = test.drop(columns=['bio1', 'bio2', 'bio3','bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 
                        'bio11','bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18','bio19'])
ytest =  test[my_variable]  


param_grid = {
        'n_estimators': [200, 400, 600, 800],
        'max_depth': [None, 5, 10, 20],
        'min_samples_split': [2, 5, 10],
        'min_samples_leaf': [1, 2, 4],
        'bootstrap': [True, False]}


train_val_splits = five_splits

X_train_val = np.concatenate([x for x,_ in train_val_splits])
y_train_val = np.concatenate([y for _,y in train_val_splits])

split_start_indices = np.cumsum([0] + [len(x) for x,_ in train_val_splits])
split_indices = np.concatenate([ i*np.ones_like(y) for i, (_, y) in enumerate(train_val_splits) ]) # split_indices[i] = 0 means sample i is part of fold0, and so on

cv = PredefinedSplit(test_fold=split_indices)

search_model = RandomizedSearchCV(estimator=etr, 
                                  param_distributions=param_grid, 
                                  n_iter=10,
                                  verbose=4, 
                                  cv=cv,
                                  refit=True, # Since refit=True, best model will be refit on the whole training+validation data
                                  random_state=42)   

search_model.fit(X_train_val, y_train_val) 

# Make predictions on the test data
y_pred = search_model.predict(Xtest)

test_metrics["r2"].append(r2_score(ytest, y_pred))
test_metrics["rmse"].append(sqrt(mean_squared_error(ytest, y_pred)))
test_metrics["mae"].append(mean_absolute_error(ytest, y_pred))
test_metrics["adjusted-r2"].append(1 - ((1 - r2_score(ytest, y_pred)) * (len(ytest) - 1) / (len(ytest) - Xtest.shape[1] - 1)))
test_metrics

search_models.append(search_model.best_params_)


# Evaluate on test data and remember the best model

print(test_metrics)



