library(readr)
library(caret)
library(randomForest)



#Loading the five folds
root <- "my/path"
fold_files <- sort(list.files(root, pattern = "*", full.names = TRUE))

#Loading the test set
test <- read.csv(paste0(root, "\\test_set.csv"), sep = ";", dec = ",")

#List containing  each fold
folds <- list()
for (i in 1:5) {
  filename <- paste0(root, "/fold", i, ".csv")
  data <- read.csv(filename, sep = ";", dec = ",", header = TRUE)
  data <- data[, !names(data) %in% c('total_taxa', 'X', 'Y', 'NPP', 'Biome')]
  folds[[i]] <- data
}

# Xtest contains all but the response variables
Xtest <- test[, !(names(test) %in% c('bio1', 'bio2', 'bio3', 'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11', 'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18', 'bio19'))]

#### ---- 1. Multiple Linear Regression

# Initialize the best_models, the search_models, and the test_metrics lists
best_models <- list()
search_models <- list()
test_metrics <- list(
  r2 = vector(),
  `adjusted-r2` = vector(),
  rmse = vector(),
  mae = vector()
)

for (j in 1:19) { # j variables to predict
  
  five_splits <- list() # store the feature data and target variable for each fold of the dataset
  for (k in 1:5) {
    fold <- folds[[k]]
    split <- list()
    split[[1]] <- fold[, !names(fold) %in% c('AI', 'bio1', 'bio2', 'bio3', 'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11', 'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18', 'bio19')]
    split[[2]] <- fold[[paste0("bio", j)]]
    five_splits[[k]] <- split
  }
  
  ytest <- test[[paste0("bio", j)]]  
  
  
  # Splits for training and validation data
  train_val_splits <- five_splits
  
  # Concatenated features and target variable for training and validation
  X_train_val <- do.call(rbind, lapply(train_val_splits, function(x) x[[1]]))
  y_train_val <- do.call(c, lapply(train_val_splits, function(x) x[[2]]))
  
  # Predefined cross-validation split
  cv <- createFolds(y_train_val, k = length(train_val_splits), returnTrain = FALSE)
  
  # Grid search with cross-validation
  search_model <- train(
    x = X_train_val,
    y = y_train_val,
    method = "lm",  # Use lm for linear regression
    trControl = trainControl(
      index = cv,
      search = "grid",
      verboseIter = TRUE
    ),
    tuneGrid = data.frame(intercept = TRUE)  # No tuning parameters for lm
  )
  
  # Fit the model on training and validation data
  search_model$finalModel
  
  # Make predictions on the test data
  y_pred <- predict(search_model, Xtest)
  test_metrics$r2 <- c(test_metrics$r2, cor(y_pred, ytest)^2)
  test_metrics$rmse <- c(test_metrics$rmse, sqrt(mean((y_pred - ytest)^2)))
  test_metrics$mae <- c(test_metrics$mae, mean(abs(y_pred - ytest)))
  test_metrics$`adjusted-r2` <- c(test_metrics$`adjusted-r2`, 1 - ((1 - cor(y_pred, ytest)^2) * (length(ytest) - 1) / (length(ytest) - ncol(Xtest) - 1)))
  
  # Store the best parameters found by the search model
  search_models <- c(search_models, search_model$bestTune)
  cat("LM model fitted for response variable bio", j, "\n")
  
}


test_metrics_lm <- data.frame(r2 = test_metrics$r2, 
                              ar2 = test_metrics$`adjusted-r2`, 
                              rmse = test_metrics$rmse, 
                              mae = test_metrics$mae, 
                              row.names = paste0("bio", 1:19))
write.csv2(test_metrics_lm, "my/path/metrics_lm.csv", row.names = T)
print(test_metrics_lm)



#### ---- 2. Elastic net regression


# Initialize the best_models, the search_models, and the test_metrics lists
best_models <- list()
search_models <- list()
test_metrics <- list(
  r2 = vector(),
  `adjusted-r2` = vector(),
  rmse = vector(),
  mae = vector()
)

for (j in 1:19) { # j variables to predict
  
  five_splits <- list() # store the feature data and target variable for each fold of the dataset
  for (k in 1:5) {
    fold <- folds[[k]]
    split <- list()
    split[[1]] <- fold[, !names(fold) %in% c('AI', 'bio1', 'bio2', 'bio3', 'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11', 'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18', 'bio19')]
    split[[2]] <- fold[[paste0("bio", j)]]
    five_splits[[k]] <- split
  }
  
  ytest <- test[[paste0("bio", j)]]  
  
  # Splits for training and validation data
  train_val_splits <- five_splits
  
  # Concatenated features and target variable for training and validation
  X_train_val <- do.call(rbind, lapply(train_val_splits, function(x) x[[1]]))
  y_train_val <- do.call(c, lapply(train_val_splits, function(x) x[[2]]))
  
  # Predefined cross-validation split
  cv <- createFolds(y_train_val, k = length(train_val_splits), returnTrain = FALSE)
  
  # Grid search with cross-validation
  search_model <- train(
    x = X_train_val,
    y = y_train_val,
    method = "glmnet",  # Use glmnet for elastic net regression
    trControl = trainControl(
      index = cv,
      search = "grid",
      verboseIter = TRUE
    ),
    tuneGrid = expand.grid(alpha = seq(0, 1, 0.5), 
                           lambda = seq(0, 1, 0.5)),
    allowParallel = TRUE 
  )
  
  
  # Fit the model on training and validation data
  search_model$finalModel
  
  # Make predictions on the test data
  y_pred <- predict(search_model, Xtest)
  test_metrics$r2 <- c(test_metrics$r2, cor(y_pred, ytest)^2)
  test_metrics$rmse <- c(test_metrics$rmse, sqrt(mean((y_pred - ytest)^2)))
  test_metrics$mae <- c(test_metrics$mae, mean(abs(y_pred - ytest)))
  test_metrics$`adjusted-r2` <- c(test_metrics$`adjusted-r2`, 1 - ((1 - cor(y_pred, ytest)^2) * (length(ytest) - 1) / (length(ytest) - ncol(Xtest) - 1)))
  
  # Store the best parameters found by the search model
  search_models <- c(search_models, search_model$bestTune)
  cat("LM model fitted for response variable bio", j, "\n")
  
  
}


test_metrics_elastic <- data.frame(r2 = test_metrics$r2, 
                                   ar2 = test_metrics$`adjusted-r2`, 
                                   rmse = test_metrics$rmse, 
                                   mae = test_metrics$mae, 
                                   row.names = paste0("bio", 1:19))
write.csv2(test_metrics_elastic, "my/path/metrics_elastic.csv", row.names = T)


#### ---- 3. Partial Least Square regression (PLS)

# Initialize the best_models, the search_models, and the test_metrics lists
best_models <- list()
search_models <- list()
test_metrics <- list(
  r2 = vector(),
  `adjusted-r2` = vector(),
  rmse = vector(),
  mae = vector()
)

for (j in 1:19) { # j variables to predict
  
  five_splits <- list() # store the feature data and target variable for each fold of the dataset
  for (k in 1:5) {
    fold <- folds[[k]]
    split <- list()
    split[[1]] <- fold[, !names(fold) %in% c('AI', 'bio1', 'bio2', 'bio3', 'bio4', 'bio5', 'bio6', 'bio7', 'bio8', 'bio9', 'bio10', 'bio11', 'bio12', 'bio13', 'bio14', 'bio15', 'bio16', 'bio17', 'bio18', 'bio19')]
    split[[2]] <- fold[[paste0("bio", j)]]
    five_splits[[k]] <- split
  }
  
  ytest <- test[[paste0("bio", j)]]  
  
  for (i in 1:5) {
    # Splits for training and validation data
    train_val_splits <- five_splits
    
    # Concatenated features and target variable for training and validation
    X_train_val <- do.call(rbind, lapply(train_val_splits, function(x) x[[1]]))
    y_train_val <- do.call(c, lapply(train_val_splits, function(x) x[[2]]))
    
    # Predefined cross-validation split
    cv <- createFolds(y_train_val, k = length(train_val_splits), returnTrain = FALSE)
    
    # Grid search with cross-validation
    search_model <- train(
      x = X_train_val,
      y = y_train_val,
      method = "pls",  # Use pls for partial least square regression
      trControl = trainControl(
        index = cv,
        search = "grid",
        verboseIter = TRUE
      ),
      tuneLength = 5,
      allowParallel = TRUE 
    )
    
    
    # Fit the model on training and validation data
    search_model$finalModel
    
    # Make predictions on the test data
    y_pred <- predict(search_model, Xtest)
    test_metrics$r2 <- c(test_metrics$r2, cor(y_pred, ytest)^2)
    test_metrics$rmse <- c(test_metrics$rmse, sqrt(mean((y_pred - ytest)^2)))
    test_metrics$mae <- c(test_metrics$mae, mean(abs(y_pred - ytest)))
    test_metrics$`adjusted-r2` <- c(test_metrics$`adjusted-r2`, 1 - ((1 - cor(y_pred, ytest)^2) * (length(ytest) - 1) / (length(ytest) - ncol(Xtest) - 1)))
    
    # Store the best parameters found by the search model
    search_models <- c(search_models, search_model$bestTune)
    cat("Model fitted for response variable bio", j, "\n")
    
    break
  }
}

test_metrics_pls <- data.frame(r2 = test_metrics$r2, 
                               ar2 = test_metrics$`adjusted-r2`, 
                               rmse = test_metrics$rmse, 
                               mae = test_metrics$mae, 
                               row.names = paste0("bio", 1:19))
write.csv2(test_metrics_pls, "my/path/metrics_pls.csv", row.names = T)



### STEP5: Predictions

# Use the function predict to run predictions 
# Here we use the list of rodents from Northwest Cameroon (Ebague et al., 2019) as an example
# see User manual and codes in Call_function for the instructions to get the faunal list

Cameroon_rodent_list <- read.csv(Cameroon_to_pred.csv)
preds_Cameroon <- predict(my_model, Cameroon_rodent_list) 


