# Building the initial dataset
# This  dataset combines climatic data from raster layers with species distribution data from shape files

# Both species distribution maps and bioclimatic variables must be downloaded in advance before running the computations.
# See the user manual (user_manual.docx) for instructions on downloading .tif and .shp files.

# Bioclimatic raster layers from Worldclim can be downloaded here: https://www.worldclim.org/data/worldclim21.html
# .tif files must be stored together in a specific folder, e.g. 'my_climate_folder'

# Species distribution shape files can be can be downloaded here: https://mol.org/species/
# .shp files must be stored together in a specific folder, e.g. 'my_rodent_folder'

# For this code to work, species variable were renamed using a specific 8 digit code...
# ... that reflects the current taxonomy. For rodent systematics, we followed Wilson et al. (2016, 2017)
# See Rodent_code.csv for the name of the species variables


########
#####---- STEP1: aggregating bioclimatic and species distribution variables -----####
##


library(raster)
library(sp)
library(rgdal)
library(readr)
library(stringr)



#a) Bioclimatic variables

# Loading the raster files
ET <- raster("path/to/ET/raster.tif")
bio1 <- raster("path/to/1st/raster.tif")
bio2 <- raster("path/to/2nd/raster.tif")
# [...] 
bio19 <- raster("path/to/19th/raster.tif")
  #In this study, we used a set of 20 climatic variables, 
  #successively called 'ET', 'bio1', 'bio2'... until 'bio19'


# Getting coordinates
r.pts <- rasterToPoints(Bio1, spatial = TRUE)
geo.prj <- "+proj=longlat +datum=WGS84 +ellps=WGS84 +towgs84=0,0,0" 
proj4string(r.pts)
r.pts@data <- data.frame(r.pts@data, X = coordinates(r.pts)[, 1],
                         Y = coordinates(r.pts)[, 2])                         
head(r.pts@data)

# Stacking the rasters together 
Clim_stack <- stack(ET, bio1, bio2, bio3, bio4, bio5, bio6, bio7, bio8, bio9, bio10, bio11, bio12, bio13, bio14, bio15, bio16, bio17, bio18, bio19)
Full_stack <- Clim_stack #One stack with the 20 cliamte variables

# Add coordinates to Full_stack
Full_stack$X <- r.pts$X[match(values(Full_stack)[[1]], values(Altitude))]
Full_stack$Y <- r.pts$X[match(values(Full_stack)[[1]], values(Altitude))]


#B) Species distribution variables

# Creating a vector with the name of the .shp files contained in 'my_rodent_folder'
all_save_files <- list.files(path = "path/to/my_rodent_folder", pattern = "(\\.shp)$", #List the files
                             ignore.case = T, all.files = T, recursive = T, full.names = F) 
all_sp_names <- str_remove(string = all_save_files, pattern = "(\\.shp)")  #Remove the extension from the list names

# Loading polygon for species i
# Note that the folder with .shp files must also contain associated .shp, .dbf and .shx files.
for (i in 1:length(all_sp_names)) 
{
  sp <- all_sp_names[i] 
  Sp_shp <- readOGR(dsn = "path/to/my_rodent_folder", layer = sp, verbose = F) #Read shapefiles
  
  # Rasterization
  Sp_raster <- rasterize(x = Sp_shp, # Occurrence coordinates
                         y = Full_stack, # Env stack grid to apply
                         field = 1, # Value to store in the raster
                         fun = function(x, ...) mean(x)) # Apply mean to keep value = 1 in case of multiple occurrences in one pixel
  Full_stack <- addLayer(Full_stack, Sp_raster)
  names(Full_stack)[i+22] <- sp   #20 = number of bioclimatic variables +2 variables X & Y
  save(Full_stack, file = "./Full_stack_data.RData")
  cat(as.character(Sys.time()), "----- Start for", sp, " - ",i,"\n") # Provides real time workflow
}

# Converting into multi-layer RasterBrick 
Full_brick <- readAll(brick(Full_stack))

# Getting data
Sp_data <- Full_brick@data@values
dim(Sp_data)

# Eventually, we want to remove the NA values for bioclimatic variables
# They correspond to pixels located in the sea
Sp_data_cleantest <- Sp_data[apply(X = !is.na(Sp_data[,1:4]), MARGIN = 1, FUN = any), ]
dim(Sp_data_cleantest) # Dimensions of the final dataset
write.csv2(x = Sp_data_cleantest, file = "my_dataset.csv") 



########
#####---- STEP2: dding supra-specific variables -----####
##

#subset of the initial dataframe that only contains  the species binary variable
my_dataset <- read.csv2("my_dataset.csv")
only_sp_df <- my_dataset[, 23:ncol(my_dataset)] #If the dataset contains 20 climatic variables, we want to select only the species variable
only_clim_df <- my_dataset[, 1:22]
  
# loop to retrieve supraspecific ranks
class1 <- list(); class2 <- list(); class3 <- list(); class4 <- list(); class_full <- list()
for (i in 1:ncol(only_sp_df)){
  class1[i] = substr(colnames(only_sp_df[i]), 1, 1)
  class2[i] = substr(colnames(only_sp_df[i]), 1, 2)
  class3[i] = substr(colnames(only_sp_df[i]), 1, 3)
  class4[i] = substr(colnames(only_sp_df[i]), 1, 5)
  class_full[i] <- colnames(only_sp_df[i])
}

# Creating a new df with  all final columns
hier_cols <- sort(unlist(unique(c(class1, class2, class3, class4, class_full))))
newdf <- data.frame(matrix(ncol = length(hier_cols), nrow = nrow(only_sp_df)))
colnames(newdf) <- hier_cols

#list with value ==1 for each row
list__1 <- apply(only_sp_df, MARGIN = 1, FUN = function(x){names(only_sp_df)[which(x == 1)]}) 

# Then we add extraspecific taxa to the list
all_taxa <- list()
for (i in 1:nrow(only_sp_df)){
  all_taxa[i] <- list(c(unique(substr(unlist(list__1[i]), 1, 1)),
                        unique(substr(unlist(list__1[i]), 1, 2)),
                        unique(substr(unlist(list__1[i]), 1, 3)),
                        unique(substr(unlist(list__1[i]), 1, 5)),
                        list__1[[i]]))
}
head(all_taxa) 

# Eventually, we set a value of 1 for the columns that match with each unit of the extraspecific vector
list_index <- lapply(all_taxa, function(taxa) which(names(newdf) %in% unlist(taxa)))
newdf[cbind(rep(seq_len(nrow(newdf)), sapply(list_index, length)), unlist(list_index))] <- 1
newdf[is.na(newdf)] <- 0

# Then we delete the unused taxonomic ranks, using the document Rodent_codes.csv
# In the second column of Rodent_codes.csv, any taxonomic ranks that are not being used or are not applicable are assigned a value of 'NA

Rodent_codes <- read.csv(".../Rodent_codes.csv", header=FALSE)
unused_taxa <- (Rodent_codes[rowSums(is.na(Rodent_codes)) > 0,])[,1]

my_df <- newdf[, -which(names(newdf) %in% unused_taxa)]

# newdf must then be appended to the initial set of bioclimatic variables, 
# but here, we want to perform  a geographical cross-validation so we need to retrieve first the XY coordinates of our rasters

final_df <- cbind(only_clim_df, my_df) 


########
#####---- STEP3: Splitting the dataset for geographical cross-validation -----####
##

XY <- cbind(final_df$X, final_df$X, final_df$bio1)

n_steps <- 30 #If we want the dataset divided by (n_steps x n_steps) blocks
# Size of the blocks must be determined to achieve a balance 
# between accommodating species with limited geographic ranges and 
# providing sufficient coverage for evaluating the models' adaptability to climatic conditions without direct analogues 

#to get a test square of L = a whose surface equals to 3/4 of the area of the training square of L = b, 
#so size(test set) = 25% of the data
#the length of b must be  L(b) = (sqrt(3)/2)*a

# Horizontal thresholds
min_Y <- min(XY$Y)
max_Y <- max(XY$Y)
vec1 <- seq(min_Y, max_Y, length.out = n_steps)
length_a <- vec1[3]-vec1[2]
length_b = (sqrt(3)/2)*length_a
vec2 <- seq(min_Y+length_b, max_Y+length_b, length.out = n_steps)
threshold_Y <- sort(c(vec1, vec2))

# Vertical thresholds
min_X <- min(XY$X)
max_X <- max(XY$X)
vecx1 <- seq(min_X, max_X, length.out = n_steps)
lengthx_a <- vecx1[2]-vecx1[1]
lengthx_b = (sqrt(3)/2)*lengthx_a
vecx2 <- seq(min_X+lengthx_b, max_X+lengthx_b, length.out = n_steps)
threshold_X <- sort(c(vecx1, vecx2))

### STEP 2: Creating random geographical blocks

copy_bdd <- final_df
copy_bdd$block <- 0
for (i in seq(from = 1, to = 60, by = 2)) {
  for (z in seq(from = 1, to = 60, by = 2)) {
    copy_bdd$block <- ifelse(
      copy_bdd$Y >= threshold_Y[i] & copy_bdd$Y < threshold_Y[i+1] 
      & copy_bdd$X >= threshold_X[z] & copy_bdd$X < threshold_X[z+1],
      sample(1:5, 1),
      copy_bdd$block
    )
  }
}

#The calculation below should approximately equals to 3 if we have a 0.75/0.25 ratio
(nrow(copy_bdd[copy_bdd$block == 1,])+  nrow(copy_bdd[copy_bdd$block == 2,])+
    nrow(copy_bdd[copy_bdd$block == 3,])+  nrow(copy_bdd[copy_bdd$block == 4,])+
    nrow(copy_bdd[copy_bdd$block == 5,]))/nrow(copy_bdd[copy_bdd$block == 0,])


fold_names <- paste0("fold", 1:5)
fold_list <- lapply(0:5, function(i) {
  subset <- copy_bdd[copy_bdd$block == i, ]
  subset <- subset[, 1:(ncol(subset) - 1)]
  assign(fold_names[i+1], subset)
})
test_set <- fold_list[[1]]
fold_list <- fold_list[-1]

#Then we export each fold and the test set separately
write.csv2(test_set, "my/path/test_set.csv", row.names = F)
write.csv2(fold1, "my/path/fold1.csv", row.names = F)
write.csv2(fold2, "my/path/fold2.csv", row.names = F)
write.csv2(fold3, "my/path/fold3.csv", row.names = F)
write.csv2(fold4, "my/path/fold4.csv", row.names = F)
write.csv2(fold5, "my/path/fold5.csv", row.names = F)

# We will use these folds for predefind split in cross-validation


